#include <boost/make_shared.hpp>
#include "list.hpp"

template<typename T>
List<T>::Element::Element(const T& _elem) : elem(_elem) {
}

template<typename T>
List<T>::Iterator::Exception::Exception(const std::string& _msg) : msg(_msg) {
}

template<typename T>
List<T>::Iterator::Exception::~Exception() throw() {
}

template<typename T>
const char* List<T>::Iterator::Exception::what() const throw() {
   return msg.c_str();
}

template<typename T>
List<T>::Iterator::Iterator() : ptr() {
}

template<typename T>
List<T>::Iterator::Iterator(WeakLink _ptr) : ptr(_ptr) {
}

template<typename T>
bool List<T>::Iterator::valid() {
   return !ptr.expired();
}

template<typename T>
T& List<T>::Iterator::operator*() {
   Link p = ptr.lock();
   if (p) {
      return p->elem;
   } else {
      throw Exception("iterator is expired");
   }
}

template<typename T>
typename List<T>::Iterator& List<T>::Iterator::operator++() {
   // prefix increment
   Link p = ptr.lock();
   if (p) {
      ptr = p->next;
   } else {
      throw Exception("iterator is expired");
   }
   return *this;
}

template<typename T>
typename List<T>::Iterator List<T>::Iterator::operator++(int unused) {
   // postfix increment
   Iterator oldit(ptr);
   Link p = ptr.lock();
   if (p) {
      ptr = p->next;
   } else {
      throw Exception("iterator is expired");
   }
   return oldit;
}

template<typename T>
typename List<T>::Iterator& List<T>::Iterator::operator--() {
   // prefix decrement
   Link p = ptr.lock();
   if (p) {
      ptr = p->prev;
   } else {
      throw Exception("iterator is expired");
   }
   return *this;
}

template<typename T>
typename List<T>::Iterator List<T>::Iterator::operator--(int unused) {
   // prefix decrement
   Iterator oldit(ptr);
   Link p = ptr.lock();
   if (p) {
      ptr = p->prev;
   } else {
      throw Exception("iterator is expired");
   }
   return oldit;
}

template<typename T>
bool List<T>::Iterator::operator==(const Iterator& other) {
   Link p1 = ptr.lock();
   Link p2 = other.ptr.lock();
   return p1 == p2;
}

template<typename T>
bool List<T>::Iterator::operator!=(const Iterator& other) {
   return !(*this == other);
}

template<typename T>
typename List<T>::Iterator List<T>::begin() {
   return Iterator(head);
}

template<typename T>
typename List<T>::Iterator List<T>::end() {
   return Iterator();
}

template<typename T>
void List<T>::push_back(const T& object) {
   Link ptr = boost::make_shared<Element>(object);
   ptr->prev = tail;
   if (head) {
      tail->next = ptr;
   } else {
      head = ptr;
   }
   tail = ptr;
}
