#ifndef LIST_H
#define LIST_H

#include <boost/shared_ptr.hpp>
#include <boost/weak_ptr.hpp>
#include <exception>
#include <string>

template <typename T>
class List {
   private:
      struct Element;
      typedef boost::shared_ptr<Element> Link;
      typedef boost::weak_ptr<Element> WeakLink;
      struct Element {
	 Element(const T& _elem);
	 T elem;
	 Link next;
	 WeakLink prev;
      };
      Link head;
      Link tail;
   public:
      class Iterator {
	 public:
	    class Exception: public std::exception {
	       public:
		  Exception(const std::string& _msg);
		  virtual ~Exception() throw();
		  virtual const char* what() const throw();
	       private:
		  std::string msg;
	    };
	    bool valid();
	    T& operator*();
	    Iterator& operator++(); // prefix increment
	    Iterator operator++(int); // postfix increment
	    Iterator& operator--(); // prefix decrement
	    Iterator operator--(int); // postfix decrement
	    bool operator==(const Iterator& other);
	    bool operator!=(const Iterator& other);
	 private:
	    friend class List;
	    Iterator();
	    Iterator(WeakLink _ptr);
	    WeakLink ptr;
      };
      Iterator begin();
      Iterator end();
      void push_back(const T& object);
};

#endif
